"""
    Data loaders for SPEC and PES.
"""

from extra_data import open_run, RunDirectory
import numpy as np

from typing import Any, Union, List, Dict

import os
import h5py
import numpy as np

import itertools

import scipy
import scipy.interpolate
from random import choices

from skimage.feature import canny
import torch
import torchvision.transforms as transforms
from collections import namedtuple
import torch.nn.functional as F

import cv2

import sys
sys.path.append("./")
sys.path.append("..")


from sklearn.datasets import fetch_olivetti_faces
from numpy.random import RandomState


class ReadPesSpec:
    """
    Read PES and SPEC data, select by unique IDs
    """

    def __init__(self, data_info):

        self.data_info = data_info

        self.rundir = self.data_info["run_directory"] 
        self.runid = self.data_info["run_number"]
        self.spec_ofset = self.data_info["spec_ofset"]

        self.run = RunDirectory(f"{self.rundir}/{self.runid}") 

        self.spec_inds = None
        self.pes_inds = None
        self.xgm_inds = None

        self.spec_tid_all = None
        self.spec_tid_all_offset = None
        self.pes_tid_all = None
        self.xgm_tid_all = None

        self.spec_tids = None
        self.pes_tids = None
        self.xgm_tids = None

        self.retvol_raw = None
        self.retvol_raw_timestamp = None
        self.min_ret_vol = None


        self.spec_raw_pe = None 
        self.spec_raw_int = None 
        self.pes_raw_dict = None  
        self.pes_data = None
        self.xgm_pressureFiltered = None  
        self.xgm_pulseen = None
        self.retvol_raw = None  
        self.raw_timestamp = None 
        self.min_ret_vol = None 


        # Output
        self.data = {"spec_raw_pe":None,
                    "spec_raw_int":None,
                    "pes_raw_dict":None,
                    "pes_data":None,
                    "xgm_pressureFiltered":None,
                    "xgm_pulseen":None,
                    "retvol_raw":None,
                    "retvol_raw_timestamp":None,
                    "min_ret_vol":None}



    def get_unq_inds(self, a, b, c):
        uniq_vals = list(set(a).intersection(b).intersection(c))
        return [a.index(x) for x in uniq_vals], [b.index(x) for x in uniq_vals], [c.index(x) for x in uniq_vals]

    # create channel names
    def gen_channel_names(self):
        channels = []
        for i in range(4):
            channels.append(f"channel_{i+1}_A")
            channels.append(f"channel_{i+1}_B")
            channels.append(f"channel_{i+1}_C")
            channels.append(f"channel_{i+1}_D")

        chs_short_name = []
        for i in range(4):
            chs_short_name.append(f"ch_{i+1}_A")
            chs_short_name.append(f"ch_{i+1}_B")
            chs_short_name.append(f"ch_{i+1}_C")
            chs_short_name.append(f"ch_{i+1}_D")

        return channels, chs_short_name


    def read_train_ids(self, debug=False):

        # Read raw SPEC Data train IDs
        self.spec_tid_all = self.run['SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output', f"data.trainId"].ndarray()
        self.spec_tid_all_offset = self.spec_tid_all + self.spec_ofset  ##### Offset correction  #####
        #print(f" spec_tid_all_offset = {len(spec_tid_all_offset)}")

        # Read raw PES Data train IDs
        self.pes_tid_all = self.run['SA3_XTD10_PES/ADC/1:network', f"digitizers.trainId"].ndarray() 
        #print(f" pes_tid_all = {len(pes_tid_all)}")


        # Read raw XGM Data train IDs
        self.xgm_tid_all = self.run['SA3_XTD10_XGM/XGM/DOOCS:output', f"data.trainId"].ndarray() 
        #print(f" xgm_tid_all = {len(xgm_tid_all)}") 

        print(f" Introduced SPEC_OFFSET = {self.spec_ofset} \n ")

        if debug:
            print(f"\n Len of train IDs:                           \n \
                spec_tid_all_offset  {len(self.spec_tid_all_offset)}    \n \
                pes_tid_all          {len(self.pes_tid_all)}            \n \
                xgm_tid_all          {len(self.xgm_tid_all)}            \n \
                    ")


            print(f"\n First 5 train IDs from all sources:         \n \
                spec_tid_all         {self.spec_tid_all[:5]}            \n \
                spec_tid_all_offset  {self.spec_tid_all_offset[:5]}     \n \
                pes_tid_all          {self.pes_tid_all[:5]}             \n \
                xgm_tid_all          {self.xgm_tid_all[:5]}             \n \
                ")


    def merge_train_ids(self, debug=True):
        # Merge Train IDs
        self.spec_inds, self.pes_inds, self.xgm_inds = self.get_unq_inds(list(self.spec_tid_all_offset), list(self.pes_tid_all), list(self.xgm_tid_all))


        print(" Len of indices for common elements in spec, pes, xgm :", len(self.spec_inds), len(self.pes_inds), len(self.xgm_inds))

        # Find common train ids
        self.spec_tids = self.spec_tid_all_offset[self.spec_inds]
        self.pes_tids = self.pes_tid_all[self.pes_inds]
        self.xgm_tids = self.xgm_tid_all[self.xgm_inds]

        if debug:
            print(f"\n First 5 indices from all sources:    \n \
                spec_inds         {self.spec_inds[:5]}          \n \
                pes_inds          {self.pes_inds[:5]}           \n \
                xgm_inds          {self.xgm_inds[:5]}           \n \
                    ")

            print(f"\n First 5 train IDs from all sources:    \n \
                spec_tids         {self.spec_tids[:5]}          \n \
                pes_tids          {self.pes_tids[:5]}           \n \
                xgm_tids          {self.xgm_tids[:5]}           \n \
                ")
        return  self.spec_inds, self.pes_inds, self.xgm_inds


    def retVol(self):
        # Read PES Data RetVol
        self.retvol_raw = self.run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.value"].ndarray()[list(self.pes_inds)]
        self.min_ret_vol = np.abs(self.retvol_raw.min())
        self.retvol_raw_timestamp = self.run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.timestamp"].ndarray()[list(self.pes_inds)]


    def read_data(self, debug=True):

        # Read raw SPEC Data
        self.spec_raw_pe= self.run['SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output', f"data.photonEnergy"].ndarray().astype(float)[list(self.spec_inds),:]
        #spec_raw_all = run_spec['SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output', f"data.intensityDistribution"].ndarray()
        self.spec_raw_int = self.run['SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output', f"data.intensityDistribution"].ndarray().astype(float)[list(self.spec_inds),:]

        # Read PES Data
        channels, _ = self.gen_channel_names()                             # create channel names
        self.pes_raw_dict = {}
        pes_raw_dict_all = {}
        for ch in channels:
            #pes_raw = run['SA3_XTD10_PES/ADC/1:network', f"digitizers.{ch}.raw.samples"].ndarray()
            #pes_raw_dict_all[ch] = run_pes['SA3_XTD10_PES/ADC/1:network', f"digitizers.{ch}.raw.samples"].ndarray()
            self.pes_raw_dict[ch] = self.run['SA3_XTD10_PES/ADC/1:network', f"digitizers.{ch}.raw.samples"].ndarray().astype(float)[list(self.pes_inds),:]
        
        # Read XGM gas pressure
        self.xgm_pressureFiltered= self.run['SA3_XTD10_XGM/XGM/DOOCS', f"pressure.pressureFiltered.value"].ndarray().astype(float)[list(self.xgm_inds)]

        # Read PES Data RetVol
        self.retVol()

        # Read XGM Pusle Energy
        self.xgm_pulseen =  self.run['SA3_XTD10_XGM/XGM/DOOCS:output', f"data.intensitySa3TD"].ndarray().astype(float)[list(self.xgm_inds)]

        if debug:
            #print(f"\n pes data channels :: \n {self.pes_raw_dict.keys()}")
            print(f"spec_raw_pe.shape{self.spec_raw_pe.shape}, \n spec_raw_int.shape{self.spec_raw_int.shape}")
            print(f"\n pes channel shape channel_1_A:: {self.pes_raw_dict['channel_1_A'].shape}")
            print("\n shape of xgm_pressureFiltered", self.xgm_pressureFiltered.shape)
            print(f"\n min_ret_vol = {self.min_ret_vol}")


        assert self.spec_raw_pe.shape, self.spec_raw_int.shape 
        assert self.pes_raw_dict['channel_1_A'].shape[0], self.spec_raw_int.shape[0]



    def get_data(self):

        # Read Train Ids
        self.read_train_ids(self.run)

        # Merge Train Ids
        self.merge_train_ids()

        # Read Data
        self.read_data()
        
        self.data["spec_raw_pe"] = self.spec_raw_pe
        self.data["spec_raw_int"] = self.spec_raw_int
        self.data["pes_raw_dict"] = self.pes_raw_dict
        self.data["xgm_pressureFiltered"] = self.xgm_pressureFiltered
        self.data["xgm_pulseen"] = self.xgm_pulseen
        self.data["retvol_raw"] = self.retvol_raw
        self.data["retvol_raw_timestamp"] = self.retvol_raw_timestamp
        self.data["min_ret_vol"] = self.min_ret_vol
        
        print("Data loaded in dataframe")

        PCS = PesChannelSelector(pes_dict=self.data, ch_names=self.data_info["channel_names"])    # Select Pes Channels
        self.pes_data = PCS.sel_pes_chs()  
        self.data["pes_data"] = self.pes_data

        print("pes_data.shape = ", self.pes_data.shape)      #  (14497, 1200)
        print("spec_data.shape = ", self.spec_raw_int.shape)    #  (14497, 1900)


        return self.data


class PesChannelSelector:
    """_Use this class to take all or part of PES channels for data preprocessing_
    """
    def __init__(self, pes_dict, ch_names="default"):

        self.pes_dict = pes_dict
        self.default_channels = ["channel_1_D", "channel_2_B", "channel_3_A", "channel_3_B", "channel_4_C", "channel_4_D"]


        if ch_names is "default":
            self.channels = self.default_channels
        elif ch_names == "all_chs":
            self.channels, _ = self.gen_channel_names()
        else:
            self.channels = ch_names   

        # Output
        self.pes_data =  np.array([])



    # create channel names
    def gen_channel_names(self):
        """
        Generate channel names for quick reference

        Returns:
            [channels], [chs_short_name]: raw channel names, channel short names
        """
        channels = []
        for i in range(4):
            channels.append(f"channel_{i+1}_A")
            channels.append(f"channel_{i+1}_B")
            channels.append(f"channel_{i+1}_C")
            channels.append(f"channel_{i+1}_D")

        chs_short_name = []
        for i in range(4):
            chs_short_name.append(f"ch_{i+1}_A")
            chs_short_name.append(f"ch_{i+1}_B")
            chs_short_name.append(f"ch_{i+1}_C")
            chs_short_name.append(f"ch_{i+1}_D")

        return channels, chs_short_name


    def get_array_from_dict(self, ch_name):
        """ Select ROI where TOF is present

        Args:
            ch_name (str): channel name

        Returns:
            Croped channel : crop channel around the TOF measurement
        """
        TOF_start = 31445 + 0
        TOF_end = TOF_start + 200
        return self.pes_dict["pes_raw_dict"][ch_name][:,TOF_start:TOF_end]


    def sel_pes_chs(self):
        """ Select PES channel

        Returns:
            pes_data: PES data concat together all channels
        """

        self.pes_data = np.concatenate([self.get_array_from_dict(ch_name) for ch_name in self.channels], axis=1) 
        
        return self.pes_data 
                  
